package src.findPeaks;

import java.io.FileWriter;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.Vector;

import src.lib.CoordHitPlus;
import src.lib.CoordIterator;

/**
 * Used to find peaks usig mapped tags from a ChIP-type experiment.
 * 
 * Takes a .coord file, interprets that reads as Single_end_tags (SETs) - and
 * extends them into their full pet fragment (xSETs).
 * 
 * This really should only be used with ELAND alignments, or something else that
 * ensures there are no ambiguous reads.
 * 
 * This needs the CoordIterator.class and the CoordHitPlus.class
 * 
 * @version 2.1 - rewritten for clarity and bug fixes
 * @author Anthony Fejes
 * 
 */

public class FindPeaks {

	/**
	 * @param args
	 */
	static Hashtable<String, Vector<CoordHitPlus>> chromo = new Hashtable<String, Vector<CoordHitPlus>>(50);

	static Hashtable<String, String> chromoNames = new Hashtable<String, String>(50);

	static String SVNID = "$Id: FindPeaks.java 32193 2007-09-10 20:28:02Z afejes $";

	static int minPeak = 0; // minimum peak height

	/*
	 * This is the length of the Fragments, includng tags and extension lengths.
	 * It can now be set at the command line as the 4th parameter.
	 * The default value should be 174, as previously chosen for other experiments.
	 */
	static int TOTAL_LENGTH = 174;

	static int totalTagsInPeaks = 0;

	static int totalPeakLength = 0;

	/*
	 * the total number of tags observed
	 */
	static int total_tags = 0;
	
	/*
	 * the total number of unique tags observed - used for filtered runs
	 */
	static int total_unique_tags = 0;

	static int longestPeak = -1;

	static int shortestPeak = Integer.MAX_VALUE;

	// minimum nuber of duplicate tags requred to call a peak saturated
	static int mindupes = 2;

	static boolean uniqueDuplicates = true;

	static double genomesize = 2.8E9;

	static int histoSize = 31;
	
	static long[] plusStrand = new long[2001];

	static long[] minusStrand = new long[2001];
	
	static String path = "";
	
	static int duplicate_tags 	= 0;
	
	static int smallest		= 100000;
	
	static int largest 		= 0;

	/*
	 * A dupe is a read that starts at the same position as another read. If the
	 * DNA has been PCRed then this is likely a PCR induced duplicate.
	 */
	static boolean allowDupes = true;

	
	/**
	 * Processes Parameters from the command line, and launches findPeaks Code
	 * 
	 * @param args
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception {
		System.out.println("FindPeaks Version: " + FindPeaks.SVNID);
		System.out.println("CoordIterator Version: " + CoordIterator.SVNID);
		System.out.println("CoordHitPlus Version: " + CoordHitPlus.SVNID);
		if (args.length < 4) {
			System.out.println("Find the peaks in a set of mapped tags, output as a .wig file!");
			usage();
		}
		String namingscheme = args[0];
		String s_dupes = args[1];
		if (s_dupes.equalsIgnoreCase("Y")) {
			allowDupes = true;
		} else if (s_dupes.equalsIgnoreCase("N")) {
			allowDupes = false;
		} else {
			System.out.println("Second argument must be [Y|N] - Y = allow duplicates to contribute to peak height, N = do not allow duplicates to contribut to peak height");
			usage();
		}
		try {
			minPeak = Integer.parseInt(args[2]);
		} catch (Exception e) {
			System.out.println("Third argument must be a height threshold cut off!!");
			usage();
		}
		try {
			TOTAL_LENGTH = Integer.parseInt(args[3]);
		} catch (Exception e) {
			System.out.println("Fourth argument must be the length of the Fragment, including tags!!");
			usage();
		}
		try {
			histoSize = Integer.parseInt(args[4]) + 1;
		} catch (Exception e) {
			System.out.println("Fifth argument must be the size of the histogram (try 30!)");
			usage();
		}

		try {
			path = args[5];
		} catch (Exception e) {
			System.out.println("Sixth argument must be the output path. This must end in a trailing slash.");
			usage();
		}
		
		
		String[] files = new String[args.length - 6];
		for (int i = 0; i < files.length; i++) {
			files[i] = args[i + 6];
		}
		findPeaks(files, namingscheme);
	}

	public static void usage() {
		System.out.println("Usage: FindPeaks NAME Y/N (allow duplicates) <ht cut off> <fragment length> <histogram size> <output directory/> <coordfiles of type 21.whatever.coord 22.wahtever.coord X.wahtever.coord>");
		System.exit(-1);
	}

	/**
	 * Loads the coords into a vector for each sequence in the database (usually
	 * chromosomes).
	 * 
	 * APF: Fragment: 
	 * 		1 = first tag 
	 * 		2 = second tag 
	 * 		E = extention in the middle
	 * 
	 * 1111111EEEEEEEEEEEEEEEEEEEEEEEEE2222222 
	 * | 	  \                       /      |
	 * start1  End1             start2    End2
	 * 
	 * Ensures that there is only a "best hit" kept.
	 * 
	 * @param file
	 *            The coord file
	 * @throws Exception
	 */
	public static Vector<CoordHitPlus> loadSingleChrCoord(String file)
			throws Exception {
		CoordIterator ci = new CoordIterator(file);
		int cnt = 0;
		Vector<CoordHitPlus> hits = new Vector<CoordHitPlus>(100000);
		System.out.print("Loading tags :");
		while (ci.hasNext()) {
			cnt++;
			if (cnt % 10000 == 0) {
				System.out.print(".");
			}
			Vector<CoordHitPlus> chp = ci.next();
			if (chp.size() == 0) {
				continue;
			}
			if (chp.size() > 1) {
				System.err.println("Warning, read has multiple hits.  Selecting only the best.");
				System.out.println("CHP: " + chp.toString());
				CoordHitPlus best = CoordIterator.getBestPositives(chp);

				// blow away some data from best...to make it use less room?
				best.sequence = null;
				best.homology = null;
				best.query = null;
				chp.clear();
				chp.add(best);
			}

			// chp should now only ever have one entry!
			for (int i = 0; i < chp.size(); i++) {
				CoordHitPlus curr = chp.elementAt(i);
				if (curr.direct == '-') {
					curr.hstart = curr.hend - (FindPeaks.TOTAL_LENGTH -1);
					curr.hend = curr.hstart + curr.qend;
					if (curr.hstart < 0 || curr.hend < 0) {
						continue;
					}
				}
				hits.add(curr);
			}
		}
		System.out.println();
		return hits;
	}


	/**
	 * Code to count duplicat tags in data set - Uses sorted data to find the
	 * full number of tags. **Assumes data is sorted in correct order**
	 * 
	 * @param hitLocations
	 * @return
	 * @author Anthony Fejes
	 */
	private static Vector<CoordHitPlus> filterDuplicateTags(CoordHitPlus[] chp) {
		Vector<CoordHitPlus> filtered = new Vector<CoordHitPlus>();
		for (int x = 0; x < chp.length-1; x++) {
			if (chp[x].hstart != chp[x+1].hstart) {
				CoordHitPlus cp = new CoordHitPlus();
				cp = chp[x];
				filtered.add(cp);
			} else {
				if (chp[x].direct != chp[x+1].direct) {
					CoordHitPlus cp = new CoordHitPlus();
					cp = chp[x];
					filtered.add(cp);
				}
			}
		}
		/*
		 * You must always add the last record. If it's not a repeat, it won't
		 * have been added above, and if it is a repeat, the ones before it
		 * won't have been added.
		 */
		filtered.add(chp[chp.length-1]);
		return filtered;
	}


	/**
	 * Function to generate an int array that represents the height at each
	 * position in a peak.
	 * 
	 * @param first_hit
	 * @param last_hit
	 * @param hits
	 * @return
	 * @author Anthony Fejes
	 */
	private static int[] generatePeakHeights(int first_hit, int last_hit, CoordHitPlus[] hits) {
		// this will store the coverage at each position in the peak
		
		int Span = TOTAL_LENGTH + (hits[last_hit].hstart - hits[first_hit].hstart);
		int offset = hits[first_hit].hstart;
		int[] coverageHeights = new int[Span];		
		for (int y = first_hit; y <= last_hit; y++) {
			for (int q = 0; q < TOTAL_LENGTH; q++) {
				coverageHeights[(hits[y].hstart - offset) + q]++;
			}
		}
		return coverageHeights;
	}

	/**
	 * Returns the maximum value stored in an unsorted int array in O(n) time
	 * 
	 * @param list
	 * @return
	 * @author Anthony Fejes
	 */
	private static int maximum(int[] list) {
		int max=0;
		for (int x = 0; x < list.length; x++ ) {
			if (list[x] > max) {
				max = list[x];
			}
		}
		return max;
	}
	
	/**
	 * peak centre returns the centre index value of the longest stretch of numbers with "value".
	 * If value is set to the maximum, this provides the "best centre" match for a peak.
	 * @param list
	 * @param value
	 * @return
	 * @author Anthony Fejes
	 */
	private static int peakCentre(int[] list, int value) {
		int cur_start = 0;
		int cur_length = 0;
		int best_start = 0;
		int best_length = 0;
		
		for (int q = 0; q < list.length; q++) {
			if (list[q] == value) {
				cur_start = q;
				cur_length = 1;
				while (((q + 1) < list.length) && (list[q + 1] == value)) {
					cur_length++;
					q++;
				}
			}
			if (cur_length > best_length) {
				best_length = cur_length;
				best_start = cur_start;
			}
		}
		//half way down the length
		//+best start, the location of the height max. 
		return (best_length/2) + best_start;
	}
	
	/**
	 * Initializes and writes headers for wig and peak files.
	 * 
	 * @param wig
	 * @param peaks
	 * @param fname
	 * @throws Exception
	 */
	private static void initialize_files(FileWriter wig, FileWriter peaks, String fname) throws Exception {
		System.out.println("Finding peaks...");

		wig.write("# FindPeaks Version: " + FindPeaks.SVNID + "\n");
		wig.write("# CoordIterator Version: " + CoordIterator.SVNID + "\n");
		wig.write("# CoordHitPlus Version: " + CoordHitPlus.SVNID + "\n");

		String ucsc_description = fname;
		if (ucsc_description.indexOf("/") > -1) {
			ucsc_description = ucsc_description.substring(ucsc_description
					.indexOf("/"));
		}
		if (ucsc_description.lastIndexOf(".") > -1) {
			ucsc_description = ucsc_description.substring(0, ucsc_description
					.lastIndexOf("."));
		}
		String simpleName = ucsc_description;
		ucsc_description += "_ht:" + minPeak + "_FL:" + TOTAL_LENGTH;
		if (allowDupes) {
			ucsc_description += "_dupe_rds_inc";
		} else {
			ucsc_description += "_no_dupes";
		}
		wig.write("track name=\"" + simpleName ); 
		if (allowDupes){
			wig.write("_duplicates");
			
		} else {
			wig.write("_no_duplicates");
		}
		wig.write("\" description=\"" + ucsc_description + "\" type=\"wiggle_0\" ");
		wig.write("color=50,50,150 yLineMark=0.0 yLineOnOff=on visibility=2 ");
		wig.write("maxHeightPixels=40:40:2 priority=1 \n");
		
	}
	
	/**
	 * Writes out a peak to a peak file.
	 * 
	 * @param peaks
	 * @param chr
	 * @param start
	 * @param end
	 * @param pk_location
	 * @param maxht
	 * @param numtags
	 * @throws Exception
	 */
	private static void writePeak(FileWriter peaks, String chr, int start,
			int end, int pk_location, int maxht, int numtags) throws Exception {
		peaks.write(chr + "\t" + start + "\t"+ end + "\t");
		peaks.write(pk_location + "\t" + maxht + "\t" + numtags + "\n");
	}
	
	
	private static int find_peak_end(int i, CoordHitPlus[] hits ) {
		
		if (i == hits.length - 1) {
			return i;
		}
		while ((i+1 < hits.length ) && ((hits[i].hstart + TOTAL_LENGTH - 1) >= hits[i + 1].hstart)) {
			i++;
		}
		return i;
	}
	
	

	private static void findPeaks(String[] files, String fname) throws Exception {
		int[] histo 		= new int[histoSize];
		Vector<Integer> stats = new Vector<Integer>();

		FileWriter wig = new FileWriter(path + fname + ".wig");
		FileWriter peaks = new FileWriter(path + fname + ".peaks");
		
		initialize_files(wig, peaks, fname);

		/*
		 * APF: This is the main loop over all of the chromosomes
		 */
		for (String chr : files) {
			System.out.print("Loading: " + chr);
			if (chr.indexOf("MT") != -1) {
				System.out.println(" - Skipping.");
				continue;
			}  else {
				System.out.print("\n");
			}
			Vector<CoordHitPlus> locations = loadSingleChrCoord(chr);
			CoordHitPlus[] hitLocations = new CoordHitPlus[locations.size()];
			hitLocations = locations.toArray(hitLocations);

			for (CoordHitPlus chp : hitLocations) {
				chp.sort_score = -chp.hstart;
			}
			Arrays.sort(hitLocations);
			
			/*
			 * Prefilter the list!  No duplicate checking needs to be done elsewhere, now 
			 */
			total_tags += hitLocations.length;
			
			if (!allowDupes) {
				int temp_tags = hitLocations.length;
				locations = filterDuplicateTags(hitLocations);
				hitLocations = new CoordHitPlus[locations.size()];
				hitLocations = locations.toArray(hitLocations);
				total_unique_tags += hitLocations.length;
				duplicate_tags += temp_tags - hitLocations.length;
			} 
			
			System.out.print("Finding peaks:");
			int cnt = 0;
			for (int i = 0; i < hitLocations.length; i++) {
				cnt++;
				if (cnt % 10000 == 0) {
					System.out.print(".");
				}
				int LastHit = find_peak_end(i, hitLocations);
				int[] coverageHeights = generatePeakHeights(i, LastHit, hitLocations);
				int maxht = maximum(coverageHeights); // the max ht of the current peak
						
				if (maxht >= minPeak) {
					int numTags = LastHit - i + 1;
					totalTagsInPeaks += numTags;
					int peakStart = hitLocations[i].hstart;
					int peakEnd = hitLocations[LastHit].hstart + FindPeaks.TOTAL_LENGTH -1;
					int peakLength = peakEnd - peakStart +1;
					if (peakLength < shortestPeak) {
						shortestPeak = peakLength;
					}
					if (peakLength > longestPeak) {
						longestPeak = peakLength;
					}
					totalPeakLength += peakLength;
					// Max Height, and peak centre code
					
					int mhdex = peakCentre(coverageHeights, maxht);  //the centre of the peak
					stats.add(maxht);
					
					wig.write("fixedStep chrom=chr" + hitLocations[LastHit].seqName + " start="
							+ peakStart + " step=1 \n");
					if (maxht > largest) {
						largest = maxht;
					}
					if (maxht < smallest) {
						smallest = maxht;
					}

					/*
					 * APF: This simply writes out the height at every point
					 * across the peak, and then writes out the peak summary.
					 */
					for (int qq = 0; qq < coverageHeights.length; qq++) {
						assert (coverageHeights[qq] != 0);
						wig.write(coverageHeights[qq] + "\n");
					}

					// write out a peak
					writePeak(peaks, hitLocations[i].seqName,
							hitLocations[i].hstart,
							(hitLocations[LastHit].hstart + TOTAL_LENGTH - 1),
							(mhdex + 1), maxht, numTags);
					
					// APF: moves you on to the next peak!  
					i = LastHit;
					//System.out.println("Moving to record: " + enddex + " and then ++");
				} 
			}
			wig.flush();
			System.out.println();
		}
		System.out.println("Processing compete.");
	
		// calculate some stats on the peaks, and write out to screen
		output(histo, stats);
		
		peaks.close();
		wig.close();
		System.out.println("Wrote to: " + path + fname + ".wig");
		System.out.println("Wrote to: " + path + fname + ".peaks");
	}

	/**
	 * This function gives k! (factorial)
	 *  
	 * @param k
	 * @return
	 * @author Matthew Bainbridge
	 */
	private static double bang(int k) {
		double k_factorial = 1;
		for (int j = 1; j <= k; j++) {
			k_factorial *= j;
		}
		return k_factorial;
	}

	/**
	 * Prints the histogram (Code Unverified)
	 * @throws Exception
	 * @author Matthew Bainbridge
	 */
	private static void printReadHisto() throws Exception {
		FileWriter fw = new FileWriter(path + "readhisto.txt");
		fw.write("index,plus strand count,minus strand count\n");
		for (int i = -1000; i <= 1000; i++) {
			fw.write(i + "," + plusStrand[i + 1000] + ","
					+ minusStrand[i + 1000] + "\n");
		}
		fw.close();
	}
	
	
	/** 
	 * Modularizing the out put portion of the program
	 *
	 */
	private static void output(int[] histo, Vector<Integer> stats) throws Exception {	
		double[] EXParr		= new double[histo.length];
		double[] cumEXP 	= new double[histo.length];
		
		double coverage = 0;
		if (!allowDupes) {
			coverage = (double) (total_unique_tags * FindPeaks.TOTAL_LENGTH);
		} else {
			coverage = (double) (total_tags * FindPeaks.TOTAL_LENGTH);
		}
		
		double lambda = coverage / genomesize;
		
		/*
		 * APF: I admit, I really don't know what value EXParr and cumEXP represent.
		 */
		for (int i = 0; i < histo.length; i++) {
			double P = Math.pow(Math.E, lambda * -1.0d) * Math.pow(lambda, i) / bang(i);
			double exp = P * genomesize;
			EXParr[i] = exp;
		}
		cumEXP[histo.length - 1] = EXParr[histo.length - 1];
		for (int i = histo.length - 2; i > -1; i--) {
			cumEXP[i] = cumEXP[i + 1] + EXParr[i];
		}
		
		int total_pk_size = 0;
		for (int i = 0; i < stats.size(); i++) {
			int ic = stats.elementAt(i);
			total_pk_size += ic;
			if (ic >= histo.length) {
				ic = histo.length - 1;
			}
			histo[ic]++;
		}
		
		printReadHisto();
		double avg = (double) total_pk_size / (double) stats.size();
		double tag_rat = 0;
		if (!allowDupes) {
			tag_rat = (double) totalTagsInPeaks / (double) total_unique_tags * 100.0d;
		} else {
			tag_rat = (double) totalTagsInPeaks / (double) total_tags * 100.0d;			
		}
		double cov_rat = (double) totalPeakLength / genomesize * 100.0d;
		double enrich = tag_rat / cov_rat;
		
		System.out.println("*******************************************");
		System.out.println("*   Summary Statistics - Reads            *");
		System.out.println("*******************************************");
		
		System.out.println("||Total Tags|" + total_tags + "|");
		if (!allowDupes) {
			System.out.println("||Unique Tags(used)|" + total_unique_tags + "|");
			System.out.println("||Duplicate tags in genome|" + duplicate_tags +"|");
		}
		System.out.println("||Tags involved in peaks|" +  totalTagsInPeaks + "|");
		System.out.printf("||Peak Coverage|" + totalPeakLength + " (%2.2f", cov_rat); 
		System.out.println("%)|");
		System.out.println();
		System.out.println();
		System.out.println("*******************************************");
		System.out.println("*   Summary Statistics - Peaks            *");
		System.out.println("*******************************************");
		System.out.println("||Num. peaks|" + stats.size() + "|");
		System.out.println("||Average ht|" + avg + "|");
		System.out.println("||Tallest|" + largest + "|");
		System.out.println("||Smallest|" + smallest + "|");
		System.out.printf ("||Peak Coverage|%d (%2.2f", totalPeakLength, cov_rat);
		System.out.println("%)|");
		System.out.println("||Longest|" + longestPeak + "|");
		System.out.println("||Shortest|" + shortestPeak + "|");
		System.out.printf ("||Enrichment|%2.2f|\n", enrich);
		System.out.println("||Lambda|" + lambda + "|");
		System.out.println();
		System.out.println();
		System.out.println("*******************************************");
		System.out.println("*   Summary Statistics - Histogram        *");
		System.out.println("*******************************************");
		System.out.println();
		System.out.println("Poisson statistics assuming a mean fragment length of "
						+ FindPeaks.TOTAL_LENGTH + " and " + genomesize + " genome.");
		System.out.println("The last poisson number in this histogram is wrong.");
		System.out.println();
		
		if (uniqueDuplicates) {
			System.out.println("Saturated peaks require all duplicates to be at unique positions.  Thus a tag seen 18 times counts as a single duplicate");
		}

		System.out.println("||Height||Count||Poisson||FDR||");

		for (int i = 0; i < histo.length; i++) {
			System.out.print("|" + i + "|" + histo[i] + "|");
			//System.out.print(dupehisto[i] + "|");
			double FDR = Double.NaN;
			if (histo[i] > 0) {
				FDR = cumEXP[i] / (double) histo[i];
			}
			System.out.printf("%4.4e|%4.4e|", EXParr[i], FDR );
			System.out.print("\n");
		}
		
	}

}
